package ANEP;

/*
 * $Id: ANEPacket.java,v 1.7 1997/11/21 21:38:01 mwh Exp $
 * $Revision: 1.7 $
 *
 * Copyright 1997 by Carl A. Gunter, Mike Hicks, Pankaj Kakkar,
 * Jonathan Moore, Scott Nettles
 *
 */

import java.io.*;
import java.net.InetAddress;

public class ANEPacket {
  protected int version;
  protected int flags;
  protected int typeId;
  
  protected int hdrLen;
  protected int pktLen;
  
  protected TLVList TLVs;
  
  protected byte [] payload;
  
  static public int PLANtypeId = 19;
  static public int ANTStypeId = 18;
  static public int ANONrouteTypeId = 1;
  
  static public int curVersion = 1;
  static public int defaultFlags = 128;
  
  /*
   *  Constructors
   */

  protected ANEPacket() {}

  public ANEPacket (int version, int flags, int typeId,
		    int hdrLen, int pktLen,
		    TLVList TLVs,
		    byte [] payload) {
    this.version = version;
    this.flags = flags;
    this.typeId = typeId;
    this.hdrLen = hdrLen;
    this.pktLen = pktLen;
    this.TLVs = TLVs;
    this.payload = payload;
  }

  protected void setDefaults() {
    this.version = curVersion;
    this.flags = defaultFlags;    
  }

  protected void setLengths() {
    this.hdrLen = TLVs.length () + 2;
    this.pktLen = hdrLen * 4 + payload.length;
  }

  public ANEPacket (int typeId, TLVList TLVs, byte [] payload) {
    setDefaults();
    this.typeId = typeId;
    this.TLVs = TLVs;
    this.payload = payload;
    setLengths();
  }

  /*
   *  Marshalling/Unmarshalling routines
   */
  
  public byte [] marshall () throws IOException {
    int i, curLen;
    TLVList temp = TLVs;
    ByteArrayOutputStream os = new ByteArrayOutputStream ();
    
    Marshalling.putAsByte (os, version);
    Marshalling.putAsByte (os, flags);
    
    Marshalling.putAsShort (os, typeId);
    Marshalling.putAsShort (os, hdrLen);
    Marshalling.putAsShort (os, pktLen);
    
    curLen = 0;
    while (curLen < hdrLen-2) {
      temp.head().marshall(os);
      curLen += temp.head ().getLength ();
      temp = temp.tail ();
    }
    
    for (i=0; i < pktLen - hdrLen*4; i++)
      Marshalling.putByte (os, payload [i]);
    
    return os.toByteArray ();
  }
  
  public static ANEPacket unmarshall (byte [] data) 
       throws IOException, BadANEPversionException {
	 
    return unmarshall(new ByteArrayInputStream (data));
  }

  public static ANEPacket unmarshall(InputStream is) 
       throws IOException, BadANEPversionException {
	 
    int i,curLen;
    TLVList TLVs = TLVList.Nil;
    byte [] payload;

    int version = Marshalling.getAsByte (is);
    
    // verify that this is in sync with our expected version
    if (version != curVersion)
      throw new BadANEPversionException(version,curVersion);
    
    // version is OK, we can continue with expected format
    int flags = Marshalling.getAsByte (is);
    int typeId = Marshalling.getAsShort (is);
    int hdrLen = Marshalling.getAsShort (is);
    int pktLen = Marshalling.getAsShort (is);
    
    curLen = 0;
    
    while (curLen < hdrLen-2) {         
      TLVs = TLVList.Cons (TLV.unmarshall (is), TLVs);
      curLen += TLVs.head ().getLength ();
    }
    
    payload = new byte [pktLen - hdrLen*4];
    for (i=0; i < pktLen - hdrLen*4; i++)
      payload [i] = Marshalling.getByte (is);
    
    return new ANEPacket (version, 
			  flags, 
			  typeId, 
			  hdrLen, 
			  pktLen, 
			  TLVs, 
			  payload);
  }

  public int getVersion () { return version; }
  public int getFlags () { return flags; }
  public int getTypeId () { return typeId; }
  public int getHdrLen () { return hdrLen; }
  public int getPktLen () { return pktLen; }
  public TLVList getTLVs () { return TLVs; }
  public byte [] getPayload () { return payload; }

  public void addTLV(TLV T) {
    TLVs = TLVList.Cons(T,TLVs);
    setLengths();
  }

  /*
   *  Printing routines
   */
  
  private String byteToString (byte val) {
    String retval="";
    int i, b;
    
    b = (int) val & 0xff;
    for (i=0; i<8; i++) {
      if ((b%2) == 1)
	retval = "1" + retval;
      else retval = "0" + retval;
      b = b >> 1;
    }
    
    return retval;
  }

  public String toString () {
    return
      ("Version = " + version +
       "\tFlags = " + byteToString ((byte)flags) +
       "\tType ID = " + typeId +"\nTLVs:\n"+TLVs);
  }

  // Code to test Marshalling

  public static void main(String args[])
       throws IOException, BadANEPversionException {

    ANEPacket b;
    SchemeAddr addr = new SchemeAddrIPv4(InetAddress.getLocalHost());
    SchemeAddr addr2 = new SchemeAddrIPv4Port(InetAddress.getLocalHost(),
					      3324,
					      SchemeAddrIPv4Port.IPv4UDP);
    TLVList TLVs = 
      TLVList.Cons(new AddrTLV(AddrTLV.SOURCE,addr),
		   TLVList.Cons(new AddrTLV(AddrTLV.DESTINATION,addr2),
				TLVList.Nil));
    byte[] payload = ("This is a test payload").getBytes();

    ByteArrayOutputStream os = new ByteArrayOutputStream ();

    ANEPacket a = new ANEPacket(PLANtypeId, TLVs, payload);
    byte[] packet = a.marshall();
    
    ByteArrayInputStream is = new ByteArrayInputStream(packet);
    b = unmarshall(is);

    System.out.println("OUT: "+a);
    System.out.println("OUT: payload = "+new String(a.getPayload())+"\n");
    System.out.println("IN: "+b);
    System.out.println("IN: payload = "+new String(a.getPayload()));
  }
}

/*
 * $Log: ANEPacket.java,v $
 * Revision 1.7  1997/11/21 21:38:01  mwh
 * Fixed marshalling to respect network byte order.  Added additional
 * classes to support ANON.
 *
 *
 */
